// worker_pool_test.h - Header for WorkerPool test functions
// Released under the MIT License.
// Copyright, 2025, by Samuel Williams.

#pragma once

#include <ruby.h>

void Init_IO_Event_WorkerPool_Test(VALUE IO_Event_WorkerPool);
