# frozen_string_literal: true

# Released under the MIT License.
# Copyright, 2021-2026, by Samuel Williams.

require_relative "event/version"
require_relative "event/support"
require_relative "event/selector"
require_relative "event/timers"
require_relative "event/native"
